#version 430 core

// Input variables
in VertexData
{
	vec3 normal_ws;
	vec3 position_ws;
	vec2 tex_coord;
	vec3 tangent_ws;
	vec3 bitangent_ws;
} inData;

// Output variables
layout (location = 0) out vec4 final_color;

// Data of the camera
layout (std140, binding = 0) uniform CameraData
{
	mat4 projection;		// Projection matrix
	mat4 projection_inv;	// Inverse of the projection matrix
	mat4 view;				// View matrix
	mat4 view_inv;			// Inverse of the view matrix
	mat3 view_it;			// Inverse of the transpose of the top-left part 3x3 of the view matrix
	vec3 eye_position;		// Position of the eye in world space
};

// Data of the object
layout (std140, binding = 2) uniform ModelData
{
	mat4 model;			// Model matrix
	mat4 model_inv;		// Inverse of the model matrix
	mat3 model_it;		// Inverse of the transpose of the top-left part 3x3 of the model matrix
};

// Object textures and other parameters
layout (binding = 0) uniform sampler2D color_tex;
layout (binding = 1) uniform sampler2D normal_tex;
layout (binding = 2) uniform sampler2D height_tex;
// How much is 'one' in height texture in tangent space (and also in model space and world space, since the units are the same)
uniform float depth_scale_ts;
// Number of samples that are used to sample parallax direction
uniform int sample_count;

// Converts the sizes between tangent space and texture space (directions stay the same), the meaning
// is: "How many units in texture coordinate space is one unit in tangent space?".
//
// We use a simplified version here. The actual value depends on the size of the object and also on
// the "density" of the texture. With objects like spheres, where the texture is more dense at the
// poles, this value changes with each vertex, and also it is not uniform (.x and .y are different).
uniform vec2 ts_to_texs = vec2(0.5);

// Data of the material
layout (std140, binding = 3) uniform MaterialData
{
	// See the C++ code for the documentation to individual attributes
	vec3 ambient;
	vec3 diffuse;
	float alpha;
	vec3 specular;
	float shininess;
} material;

// Data of the light
struct PhongLight
{
	// See the C++ code for the documentation to individual attributes
	vec4 position;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
	vec3 spot_direction;
	float spot_exponent;
	float spot_cos_cutoff;
	float atten_constant;
	float atten_linear;
	float atten_quadratic;
};
layout (std140, binding = 1) uniform PhongLightsData
{
	vec3 global_ambient_color;
	int lights_count;
	PhongLight lights[8];
};

// Evaluates the lighting of one Phong light. The implementation is at the end of the file
void EvaluatePhongLight(in PhongLight light, out vec3 amb, out vec3 dif, out vec3 spe, in vec3 normal, in vec3 position, in vec3 eye, in float shininess);

//-----------------------------------------------------------------------

void main()
{
	// Transforms a vector in tangent space to world space
	mat3 ts_to_ws = mat3(normalize(inData.tangent_ws), normalize(inData.bitangent_ws), normalize(inData.normal_ws));
	// Transforms a vector in world space to tangent space. It is the inverse of ts_to_ws, but since
	// the matrix is orthogonal, we may use transpose instead
	mat3 ws_to_ts = transpose(ts_to_ws);

	// Task 1: Implement parallax occlusion mapping
	//		Hints:
	//			Use 'ts_to_ws' to transform directions from tangent space to world space.
	//			Use 'ws_to_ts' to transform directions from world space to tangent space (e.g. ray direction).
	//			Use 'ts_to_texs' to transform directions from tangent space to texture space (e.g. parallax).
	//			Depth scale is in 'depth_scale_ts' variable, number of samples in 'sample_count' variable.
	//			Height texture is in 'height_tex', do not forget to subtract it from one (in this height texture,
	//				1 is the surface and 0 is below the surface).
	//			Use 'textureLod(tex, coord, 0)' to sample this height texture.
	// Task 2: Improve the computation of the intersection

	// Task 1: Compute the final fragment texture coordinate, after parallax offset
	vec2 final_tex_coord = inData.tex_coord;
	
	// Task 1: Compute the final fragment position, after parallax offset
	vec3 final_position_ws = inData.position_ws;

	// Task 3: Adjust fragment's depth
	//gl_FragDepth = ...

	// Sample the color texture using the final texture coordinate
	vec3 color_tex_val = texture(color_tex, final_tex_coord).rgb;

	// Normal mapping
	vec3 normal_tex_val = texture(normal_tex, final_tex_coord).rgb * 2.0 - 1.0;
	vec3 N = normalize(ts_to_ws * normal_tex_val);

	// Compute the lighting
	vec3 Eye = normalize(eye_position - final_position_ws);

	vec3 amb = global_ambient_color;
	vec3 dif = vec3(0.0);
	vec3 spe = vec3(0.0);

	for (int i = 0; i < lights_count; i++)
	{
		// Task 4: Implement self-shadowing
		float light_intensity = 1.0;

		vec3 a, d, s;
		EvaluatePhongLight(lights[i], a, d, s, N, final_position_ws, Eye, material.shininess);
		amb += a;
		dif += d * light_intensity;
		spe += s * light_intensity;
	}

	// Compute the material
	vec3 mat_ambient = material.ambient * color_tex_val;
	vec3 mat_diffuse = material.diffuse * color_tex_val;
	vec3 mat_specular = material.specular;

	// Compute the final color
	vec3 final_light = mat_ambient * amb + mat_diffuse * dif + mat_specular * spe;
	
	final_color = vec4(final_light, 1.0);
}

//-----------------------------------------------------------------------

// Evaluates the lighting of one Phong light
// light	.. [in] parameters of the light that is evaluated
// amb		.. [out] result, ambient part
// dif		.. [out] result, diffuse part
// spe		.. [out] result, specular part
// norm		.. [in] surface normal, in the world coordinates
// pos		.. [in] surface position, in the world coordinates
// eye		.. [in] direction from the surface to the eye, in the world coordinates
// shi		.. [in] shininess of the material
void EvaluatePhongLight(in PhongLight light, out vec3 amb, out vec3 dif, out vec3 spe, in vec3 norm, in vec3 pos, in vec3 eye, in float shi)
{
	vec3 L_not_normalized = light.position.xyz - pos * light.position.w;
	vec3 L = normalize(L_not_normalized);
	vec3 H = normalize(L + eye);

	// Calculate basic Phong factors
	float Iamb = 1.0;
	float Idif = max(dot(norm, L), 0.0);
	float Ispe = (Idif > 0.0) ? pow(max(dot(norm, H), 0.0), shi) : 0.0;

	// Calculate spot light factor
	if (light.spot_cos_cutoff != -1.0)
	{
		// This is a spot light
		float spot_factor;

		float spot_cos_angle = dot(-L, light.spot_direction);
		if (spot_cos_angle > light.spot_cos_cutoff)
		{
			spot_factor = pow(spot_cos_angle, light.spot_exponent);
		}
		else spot_factor = 0.0;

		Iamb *= 1.0;
		Idif *= spot_factor;
		Ispe *= spot_factor;
	}

	// Calculate attenuation
	if (light.position.w != 0.0)
	{
		// This is a point light / spot light

		float distance_from_light = length(L_not_normalized);
		float atten_factor =
			light.atten_constant +
			light.atten_linear * distance_from_light + 
			light.atten_quadratic * distance_from_light * distance_from_light;
		atten_factor = 1.0 / atten_factor;

		Iamb *= atten_factor;
		Idif *= atten_factor;
		Ispe *= atten_factor;
	}

	amb = Iamb * light.ambient;
	dif = Idif * light.diffuse;
	spe = Ispe * light.specular;
}